"""Rodgers 2000 temperature profile."""
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.data.fittable import fitparam
from taurex.output import OutputGroup

from .tprofile import TemperatureProfile


class Rodgers2000(TemperatureProfile):
    """Layer-by-layer temperature introduced in Rodgers et al (2000)

    Inverse Methods for Atmospheric Sounding (equation 3.26).
    Featured in NEMESIS code (Irwin et al., 2008,
    J. Quant. Spec., 109, 1136 (equation 19)
    Used in all Barstow et al. papers.
    """

    def __init__(
        self,
        temperature_layers: t.Optional[npt.ArrayLike] = None,
        correlation_length: t.Optional[float] = 5.0,
        covariance_matrix: t.Optional[npt.NDArray[np.float64]] = None,
    ) -> None:
        """Initialize Rodgers 2000 temperature profile.


        Parameters
        ----------
        temperature_layers : :obj:`list`
            Temperature in Kelvin per layer of pressure

        correlation_length : float
            In scaleheights, Line et al. 2013 sets this to 7, Irwin et al sets
            this to 1.5 may be left as free and Pressure dependent parameter later.

        covariance_matrix : :obj:`array` , optional
            User can supply their own covaraince matrix

        """
        super().__init__("Rodgers2000")

        self._tp_corr_length = correlation_length
        self._covariance = covariance_matrix
        self.temperature_layers = np.array(temperature_layers)
        self.generate_temperature_fitting_params()

    def gen_covariance(self) -> npt.NDArray[np.float64]:
        """Generate the covariance matrix if None is supplied."""
        h = self._tp_corr_length
        pres_prof = self.pressure_profile

        return np.exp(
            -1.0 * np.abs(np.log(pres_prof[:, None] / pres_prof[None, :])) / h
        )

    def correlate_temp(self, cov_mat: npt.NDArray[np.float64]) -> np.float64:
        """Correlate the temperature profile using the covariance matrix."""
        cov_mat_sum = np.sum(cov_mat, axis=0)
        weights = cov_mat[:, :] / cov_mat_sum[:, None]
        return weights.dot(self.temperature_layers)

    @property
    def profile(self) -> npt.NDArray[np.float64]:
        """Returns a temperature profile."""
        cov_mat = self._covariance
        if cov_mat is None:
            cov_mat = self.gen_covariance()

        return self.correlate_temp(cov_mat)

    @fitparam(
        param_name="correlation_length",
        param_latex="$C_{L}$",
        default_fit=False,
        default_bounds=[1.0, 10.0],
    )
    def correlationLength(self) -> float:  # noqa: N802
        """Correlation length in scale heights."""
        return self._tp_corr_length

    @correlationLength.setter
    def correlationLength(self, value: float) -> None:  # noqa: N802
        """Correlation length in scale heights."""
        self._tp_corr_length = value

    def generate_temperature_fitting_params(self) -> None:
        """Generates the temperature fitting parameters

        Parameters are generated for each layer of the
        atmosphere For a 4 layer atmosphere the fitting parameters generated
        are ``T_0``, ``T_1``, ``T_2`` and ``T_3``
        """

        bounds = [1e5, 1e3]
        for idx, _ in enumerate(self.temperature_layers):
            point_num = idx + 1
            param_name = f"T_{point_num}"
            param_latex = "$T_{%i}$" % point_num

            def read_point(self, idx=idx):
                return self.temperature_layers[idx]

            def write_point(self, value, idx=idx):
                self.temperature_layers[idx] = value

            read_point.__doc__ = f"""Temperature at layer {point_num} in Kelvin"""

            fget_point = read_point

            fset_point = write_point
            default_fit = False
            self.add_fittable_param(
                param_name,
                param_latex,
                fget_point,
                fset_point,
                "linear",
                default_fit,
                bounds,
            )

    def write(self, output: OutputGroup) -> OutputGroup:
        """Write Rodgers 2000 temperature profile to output group."""
        temperature = super().write(output)

        cov_mat = self._covariance
        if cov_mat is None:
            cov_mat = self.gen_covariance()

        temperature.write_array("covariance_matrix", cov_mat)
        temperature.write_array("temperature_layers", self.temperature_layers)
        temperature.write_scalar("correlation_length", self._tp_corr_length)
        return temperature

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        """Return all input keywords."""
        return (
            "rodgers",
            "rodgers2010",
        )

    BIBTEX_ENTRIES = [
        """
        @MISC{rodger_retrievals,
            author = {{Rodgers}, Clive D.},
                title = "{Inverse Methods for Atmospheric
                Sounding - Theory and Practice}",
        howpublished = {Inverse Methods for Atmospheric Sounding - Theory
        and Practice. Series: Series on Atmospheric Oceanic and Planetary Physics},
                year = "2000",
                month = "Jan",
                doi = {10.1142/9789812813718},
            adsurl = {https://ui.adsabs.harvard.edu/abs/2000SAOPP...2.....R},
            adsnote = {Provided by the SAO/NASA Astrophysics Data System}
        }
        """,
    ]



"""Module for instrument and noise models."""
from .instrument import Instrument
from .instrumentfile import InstrumentFile
from .snr import SNRInstrument

__all__ = ["Instrument", "InstrumentFile", "SNRInstrument"]



"""Rayleigh scattering contribution."""
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.model import OneDForwardModel

from .contribution import Contribution


class RayleighContribution(Contribution):
    """Computes contribution from Rayleigh scattering."""

    def __init__(self):
        """Initialise contribution."""
        super().__init__("Rayleigh")

    def prepare_each(
        self, model: OneDForwardModel, wngrid: npt.NDArray[np.float64]
    ) -> t.Generator[t.Tuple[str, npt.NDArray[np.float64]], None, None]:
        """Compute opacity due to rayleigh scattering.

        Scattering is weighted by the mixing ratio of the gas
        from chemistry.

        Parameters
        ----------
        model: :class:`~taurex.model.model.ForwardModel`
            Forward model

        wngrid: :obj:`array`
            Wavenumber grid

        Yields
        ------
        component: :obj:`tuple` of type (str, :obj:`array`)
            Name of scattering molecule and the weighted rayeligh opacity.


        """
        from taurex.util.scattering import rayleigh_sigma_from_name

        self._ngrid = wngrid.shape[0]
        self._nmols = 1
        self._nlayers = model.nLayers
        molecules = list(model.chemistry.activeGases) + list(
            model.chemistry.inactiveGases
        )

        for gasname in molecules:
            if np.max(model.chemistry.get_gas_mix_profile(gasname)) == 0.0:
                continue
            sigma = rayleigh_sigma_from_name(gasname, wngrid)

            if sigma is not None:
                final_sigma = (
                    sigma[None, :]
                    * model.chemistry.get_gas_mix_profile(gasname)[:, None]
                )
                self.sigma_xsec = final_sigma
                yield gasname, final_sigma

    @classmethod
    def input_keywords(cls) -> t.Tuple[str]:
        """Return list of keywords for contribution."""
        return ("Rayleigh",)

    BIBTEX_ENTRIES = [
        """
        @book{cox_allen_rayleigh,
        title={Allen’s astrophysical quantities},
        author={Cox, Arthur N},
        year={2015},
        publisher={Springer}
        }
        """
    ]



"""An object that globally caches variables."""

import typing as t

from taurex.log import Logger

from .singleton import Singleton


class GlobalCache(Singleton):
    """Allows for the storage of global variables."""

    def init(self):
        self.variable_dict: t.Dict[str, t.Any] = {}
        self.log = Logger("GlobalCache")

    def __getitem__(self, key: str) -> t.Any:
        """Get a variable from the cache."""
        return self.variable_dict.get(key, None)

    def __setitem__(self, key: str, value: t.Any) -> t.Any:
        """Set a variable in the cache.

        Parameters
        ----------
        key : str
            The key to store the variable under.

        value : Any
            Value to store.

        """
        self.variable_dict[key] = value



"""Base temperature class."""
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.data.citation import Citable
from taurex.data.fittable import Fittable, derivedparam
from taurex.log import Logger
from taurex.output import OutputGroup
from taurex.output.writeable import Writeable
from taurex.planet import Planet


class TemperatureProfile(Fittable, Logger, Writeable, Citable):
    """Defines temperature profile for an atmosphere.

    *Abstract Class*

    Must define:

    - :func:`profile`

    """

    def __init__(self, name: str):
        """Initialize temperature class.

        Parameters
        ----------
        name : str
            Name used in logging

        """
        Logger.__init__(self, name)
        Fittable.__init__(self)

    def initialize_profile(
        self,
        planet: t.Optional[Planet] = None,
        nlayers: t.Optional[int] = 100,
        pressure_profile: t.Optional[npt.NDArray] = None,
    ):
        """Initializes the profile.

        Parameters
        ----------
        planet: :class:`~taurex.data.planet.Planet`

        nlayers: int
            Number of layers in atmosphere

        pressure_profile: :obj:`array`
            Pressure at each layer of the atmosphere

        """
        self.nlayers = nlayers
        self.nlevels = nlayers + 1
        self.pressure_profile = pressure_profile
        self.planet = planet

    @property
    def profile(self) -> npt.NDArray[np.float64]:
        """Temperature profile at each layer.

        Must return a temperature profile at each layer of the atmosphere

        Returns
        -------
        temperature: :obj:`array`
            Temperature in Kelvin
        """
        raise NotImplementedError

    def write(self, output: OutputGroup) -> OutputGroup:
        """Write temperature profile to output."""
        temperature = output.create_group("Temperature")
        temperature.write_string("temperature_type", self.__class__.__name__)

        return temperature

    @derivedparam(param_name="avg_T", param_latex="$\\bar{T}$", compute=False)
    def averageTemperature(self) -> float:  # noqa: N802
        """Average temperature across all layers."""
        return np.mean(self.profile)

    @classmethod
    def input_keywords(cls) -> t.Tuple[str, ...]:
        """Return all input keywords."""
        raise NotImplementedError



"""Base class for computing opacities."""
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.cache import GlobalCache
from taurex.core import Citable
from taurex.log import Logger


class Opacity(Logger, Citable):
    """This is the base class for computing opactities."""

    @classmethod
    def discover(cls) -> t.List[t.Tuple[str, t.Any]]:
        """Class method, used to discover molecular opacities.

        Should return opacities that are available in the current
        environment.

        """
        raise NotImplementedError

    @classmethod
    def priority(cls) -> int:
        """Priority of the opacity during discovery."""
        return 100

    def __init__(self, name: str) -> None:
        """Initialize opacity.

        Parameters
        ----------
        name: str
            Name of opacity for logging.

        """
        super().__init__(name)

    @property
    def resolution(self) -> float:
        """Resolution of opacity."""
        raise NotImplementedError

    @property
    def moleculeName(self) -> str:  # noqa: N802
        """Name of molecule."""
        raise NotImplementedError

    @property
    def wavenumberGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        raise NotImplementedError

    @property
    def temperatureGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        raise NotImplementedError

    @property
    def pressureGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        raise NotImplementedError

    def compute_opacity(
        self,
        temperature: float,
        pressure: float,
        wngrid: t.Optional[npt.NDArray[np.float64]] = None,
    ) -> npt.NDArray[np.float64]:
        """Must return in units of cm2."""
        raise NotImplementedError

    def opacity(
        self,
        temperature: float,
        pressure: float,
        wngrid: t.Optional[npt.NDArray[np.float64]] = None,
    ) -> npt.NDArray[np.float64]:
        """Compute the opacity for a given temperature and pressure.

        Parameters
        ----------
        temperature: float
            Temperature in K
        pressure: float
            Pressure in Pa
        wngrid: :obj:`array`, optional
            Wavenumber grid to restrict to.

        Returns
        -------
        :obj:`array`
            Opacity array


        """
        if wngrid is None:
            wngrid_filter = slice(None)
        else:
            wngrid_filter = np.where(
                (self.wavenumberGrid >= wngrid.min())
                & (self.wavenumberGrid <= wngrid.max())
            )[0]

        orig = self.compute_opacity(temperature, pressure, wngrid_filter)

        if wngrid is None or np.array_equal(
            self.wavenumberGrid.take(wngrid_filter), wngrid
        ):
            return orig
        else:
            hold_method = GlobalCache()["opacity_hold"]
            if hold_method:
                return np.interp(wngrid, self.wavenumberGrid[wngrid_filter], orig)
            else:
                return np.interp(
                    wngrid,
                    self.wavenumberGrid[wngrid_filter],
                    orig,
                    left=0.0,
                    right=0.0,
                )

    def opacityCitation(self) -> t.List[str]:  # noqa: N802
        """Citation for the specific molecular opacity.


        Returns
        -------
        list of str:
            List of string with reference information

        """
        return []



"""Base Gas class."""
import typing as t

import numpy as np
import numpy.typing as npt

from taurex.data.citation import Citable
from taurex.data.fittable import Fittable
from taurex.log import Logger
from taurex.output import OutputGroup
from taurex.output.writeable import Writeable


class Gas(Fittable, Logger, Writeable, Citable):
    """This class is a base for a single molecule or gas.
    Its used to describe how it mixes at each layer and combined
    with
    :class:`~taurex.data.profile.chemistry.taurexchemistry.TaurexChemistry`
    is used to build a chemical profile of the planets atmosphere.
    Requires implementation of:

    - func:`~mixProfile`

    """

    def __init__(self, name: str, molecule_name: str) -> None:
        """Initialize base gas class.

        Parameters
        ----------
        name: str
            Identifier for class.
        molecule_name: str
            Name of molecule


        """
        Logger.__init__(self, name)
        Fittable.__init__(self)
        self._molecule_name = molecule_name
        self.mix_profile = None

    @property
    def molecule(self) -> str:
        """
        Returns
        -------
        molecule_name: str
            Name of molecule
        """
        return self._molecule_name

    @property
    def mixProfile(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Return mix profile for each layer.
        **Requires implementation**

        Should return mix profile of molecule/gas at each layer

        Returns
        -------
        mix: :obj:`array`
            Mix ratio for molecule at each layer

        """
        raise NotImplementedError

    def initialize_profile(
        self,
        nlayers: t.Optional[int] = None,
        temperature_profile: t.Optional[npt.NDArray[np.float64]] = None,
        pressure_profile: t.Optional[npt.NDArray[np.float64]] = None,
        altitude_profile: t.Optional[npt.NDArray[np.float64]] = None,
    ) -> None:
        """Initializes and computes mix profile.

        Parameters
        ----------
        nlayers: int
            Number of layers in atmosphere

        temperature_profile: :obj:`array`
            Temperature profile of atmosphere in K. Length must be
            equal to ``nlayers``

        pressure_profile: :obj:`array`
            Pressure profile of atmosphere in Pa. Length must be
            equal to ``nlayers``

        altitude_profile: :obj:`array`
            Altitude profile of atmosphere in m. Length must be
            equal to ``nlayers``

        """
        pass

    def write(self, output: OutputGroup) -> OutputGroup:
        """
        Writes class and arguments to file

        Parameters
        ----------
        output: :class:`~taurex.output.output.Output`

        """
        gas_entry = output.create_group(self.molecule)
        gas_entry.write_string("gas_type", self.__class__.__name__)
        gas_entry.write_string("molecule_name", self._molecule_name)
        return gas_entry



"""Taurex logging module."""
import io
import logging
import typing as t

logging.getLogger("taurex").addHandler(logging.NullHandler())
"""Root logger for taurex"""

root_logger = logging.getLogger("taurex")


def setup_log(name: str) -> logging.Logger:
    """Build a logger class for the given name.

    Parameters
    ----------
    name : str
        Name of logger

    Returns
    -------
    logging.Logger
        Logger class

    """
    return logging.getLogger(name)


class TauRexHandler(logging.StreamHandler):
    """Logging Handler for Taurex 3.

    Prevents other MPI threads from writing to log
    unless they are in trouble (>=ERROR)

    Parameters
    ----------
    stream : stream-object , optional
        Stream to write to otherwise defaults to ``stderr``

    """

    def __init__(self, stream: io.IOBase = None) -> None:
        from taurex.mpi import get_rank

        super().__init__(stream=stream)

        self._rank = get_rank()

    def emit(self, record: logging.LogRecord) -> None:
        # print(record)
        if self._rank == 0 or record.levelno >= logging.ERROR:
            # msg = '[{}] {}'.format(self._rank,record.msg)
            # record.msg = msg
            return super().emit(record)
        else:
            pass


class Loggable:
    """Base class for loggable objects."""

    def __init__(self, name: t.Optional[str] = None) -> None:
        """Initialise logger."""
        name = name or self.__class__.__name__
        self._logger = setup_log(f"taurex.{name}")

    def info(self, message: str, *args: t.Any, **kwargs: t.Any) -> None:
        """See :class:`logging.Logger`."""
        self._logger.info(message, *args, **kwargs)

    def warning(self, message: str, *args: t.Any, **kwargs: t.Any) -> None:
        """See :class:`logging.Logger`."""
        self._logger.warning(message, *args, **kwargs)

    def debug(self, message: str, *args: t.Any, **kwargs: t.Any) -> None:
        """See :class:`logging.Logger`."""
        import inspect

        frame = inspect.currentframe()

        new_message = message
        if frame is not None:
            f_back = frame.f_back
            if f_back is not None:
                f_code = f_back.f_code
                new_message = f"In: {f_code.co_name}() - {message}"

        self._logger.debug(new_message, *args, **kwargs)

    def error(self, message: str, *args: t.Any, **kwargs: t.Any) -> None:
        """See :class:`logging.Logger`."""
        self._logger.error(message, *args, **kwargs)

    def critical(self, message: str, *args: t.Any, **kwargs: t.Any) -> None:
        """See :class:`logging.Logger`."""
        self._logger.critical(message, *args, **kwargs)

    def error_and_raise(
        self, exception: t.Type[Exception], message: str, *args: t.Any, **kwargs: t.Any
    ) -> Exception:
        """Print error message and raises exception."""
        self._logger.error(message, *args, **kwargs)
        raise exception(message, *args, **kwargs)



"""Ktables loaded from HDF5."""
import pathlib
import typing as t

import h5py
import numpy as np
import numpy.typing as npt

from taurex.types import PathLike
from taurex.util import sanitize_molecule_string

from ..interpolateopacity import InterpModeType, InterpolatingOpacity
from .ktable import KTable


class HDF5KTable(KTable, InterpolatingOpacity):
    """
    This is the base class for computing opactities using correlated k tables
    """

    @classmethod
    def discover(cls) -> t.List[t.Tuple[str, t.Tuple[pathlib.Path, str]]]:
        """Discover opacities from hdf5 files in path."""
        from taurex.cache import GlobalCache

        path = GlobalCache()["ktable_path"]
        if path is None:
            return []
        path = pathlib.Path(path)

        files = list(path.glob("*.h5")) + list(path.glob("*.hdf5"))

        discovery = []

        interp = GlobalCache()["xsec_interpolation"] or "linear"

        for f in files:
            splits = f.stem.split("_")
            mol_name = sanitize_molecule_string(splits[0])

            discovery.append((mol_name, [f, interp]))

        return discovery

    def __init__(
        self,
        filename: PathLike,
        interpolation_mode: t.Optional[InterpModeType] = "linear",
        in_memory: t.Optional[bool] = True,
    ) -> None:
        """Initialize HDF5KTable.

        Parameters
        ----------
        filename : PathLike
            Path to file.
        interpolation_mode : t.Optional[InterpModeType], optional
            Interpolation mode, by default "linear".
        in_memory : t.Optional[bool], optional
            Load into memory, by default True.

        Raises
        ------
        FileNotFoundError
            If file not found.


        """
        filename = pathlib.Path(filename)
        self._molecule_name = sanitize_molecule_string(filename.stem.split("_")[0])
        super().__init__(
            f"HDF5Ktable:{self._molecule_name}",
            interpolation_mode=interpolation_mode,
        )

        if not filename.exists():
            raise FileNotFoundError(f"Could not find {filename}")

        self._filename = filename
        self._spec_dict = None

        self._resolution = None
        self.in_memory = in_memory
        self._load_hdf_file(filename)

    @property
    def moleculeName(self) -> str:  # noqa: N802
        """Molecule name."""
        return self._molecule_name

    @property
    def xsecGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Cross section grid."""
        return self._xsec_grid

    def _load_hdf_file(self, filename: pathlib.Path):
        import astropy.units as u

        # Load the hdf file
        self.info(f"Loading opacity from {filename}")

        self._spec_dict = h5py.File(filename, "r")

        self._wavenumber_grid: npt.NDArray[np.float64] = self._spec_dict[
            "bin_centers"
        ].copy()
        self._ngauss: int = self._spec_dict["ngauss"][()]
        self._temperature_grid: npt.NDArray[np.float64] = self._spec_dict["t"].copy()

        pressure_units: npt.NDArray[np.float64] = self._spec_dict["p"].attrs["units"]
        try:
            p_conversion = u.Unit(pressure_units).to(u.Pa)
        except u.UnitConversionError:
            p_conversion = u.Unit(pressure_units, format="cds").to(u.Pa)

        self._pressure_grid: npt.NDArray[np.float64] = (
            self._spec_dict["p"].copy() * p_conversion
        )

        if self.in_memory:
            self._xsec_grid = self._spec_dict["kcoeff"].copy()
        else:
            self._xsec_grid = self._spec_dict["kcoeff"]
        self._weights = self._spec_dict["weights"].copy()

        self._min_pressure = self._pressure_grid.min()
        self._max_pressure = self._pressure_grid.max()
        self._min_temperature = self._temperature_grid.min()
        self._max_temperature = self._temperature_grid.max()
        self.clean_molecule_name()
        if self.in_memory:
            self._spec_dict.close()

    def clean_molecule_name(self) -> None:
        """Clean molecule name."""
        splits = self.moleculeName.split("_")
        self._molecule_name = splits[0]

    @property
    def wavenumberGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Wavenumber grid."""
        return self._wavenumber_grid

    @property
    def temperatureGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Temperature grid."""
        return self._temperature_grid

    @property
    def pressureGrid(self) -> npt.NDArray[np.float64]:  # noqa: N802
        """Pressure grid."""
        return self._pressure_grid

    @property
    def resolution(self) -> float:
        """Resolution."""
        return self._resolution

    @property
    def weights(self) -> npt.NDArray[np.float64]:
        """Weights."""
        return self._weights
